#include "Ogre.h"
//
using namespace Ogre;

#include "OgreMax/OgreMaxScene.hpp"
#include "OgrePCZSceneManager.h"
#include "HRFrameListener.h"

void createCamera();
void createCameraLight();
void addResourceLocations();
void addViewports();
void createFrameListener();
void loadDotScene();
void printNodeChildren(FILE* f, Ogre::SceneNode* scnNode, int iIdentLvl);

SceneManager* g_pSceneMgr = NULL;
//
SceneNode* g_pCamNode = NULL;
Camera* g_pCamera = NULL;
//
RenderWindow* g_pWindow = NULL;
HRFrameListener* g_frameListnr = NULL;

OgreMax::OgreMaxScene g_OgreMaxScene;

int main()
{
	// Don't need to keep the pointer as we can always getSingleton() later
	new Root(/*"plugins.cfg", "ogre.cfg", "ogre.log" - implied*/);

	//if(!(Root::getSingleton().restoreConfig()))
		if(!(Root::getSingleton().showConfigDialog()))
			return -1;

	g_pWindow = Root::getSingleton().initialise(true, "My Render Window");
//	RenderWindow* pWindow = Root::getSingleton().getAutoCreatedWindow(); // another way to get the same pointer

	// Create the SceneManager, in this case a generic one
	g_pSceneMgr = Root::getSingleton().createSceneManager("PCZSceneManager", "MyPCZSceneMgr");
	// initialize the scene manager using terrain as default zone
	((PCZSceneManager*)g_pSceneMgr)->init("ZoneType_Default");
//	g_pSceneMgr->showBoundingBoxes(true);

	createCamera();
	addViewports();
	//
	addResourceLocations();
	ResourceGroupManager::getSingleton().initialiseAllResourceGroups();
	ResourceGroupManager::getSingleton().loadResourceGroup("General"); // Not mandatory - if not called, resources are loaded on first use
	createFrameListener();
	//
	createCameraLight();
	loadDotScene();

	FILE* f = fopen("scenetree.txt", "wt");
	printNodeChildren(f, g_pSceneMgr->getRootSceneNode(), 0);
	fclose(f);

	Root::getSingleton().startRendering();
}

void createCamera()
{
	// Create the camera
	g_pCamera = g_pSceneMgr->createCamera("MainCam");

	// normally you would calculate this from the size of the viewport
	g_pCamera->setAspectRatio(1.33333f);
	// 30 degrees will give more of a long, telescopic view
	g_pCamera->setFOVy(Degree(90.0f));
	g_pCamera->setNearClipDistance(0.25f);
	g_pCamera->setFarClipDistance(100.0f);
	//
	
	// NEW: create a node for the camera and control that instead of camera directly.
	// We do this because PCZSceneManager requires camera to have a parent node 
	g_pCamNode = g_pSceneMgr->getRootSceneNode()->createChildSceneNode("CamNode");
	// attach the camera to the node
	g_pCamNode->attachObject(g_pCamera);

	// Position camera somewhere in the first room
	g_pCamNode->setPosition(Vector3(0, 1.0f, 2.0f));

	// Update bounds for camera
	g_pCamNode->_updateBounds();
}

void addViewports()
{
	// assume window is a valid pointer to an existing render window, and camera is
	// a valid pointer to an existing camera instance
	Viewport* pViewportMain = NULL;
	// second parameter is z-order, remaining (default) params are position and size, respectively
	pViewportMain = g_pWindow->addViewport(g_pCamera, 0);

	// create a smaller viewport on top, in the lower right corner center, 25% of main vp size
	// Strangely, this doesn't work with PCZ SM, the new viewport is just cleared to background color, that's all

	//Viewport* pViewportSmall = NULL;
	//pViewportSmall = g_pWindow->addViewport(g_pCamera, 1, 0.75f, 0.75f, 0.25f, 0.25f);

	//pViewportSmall->setOverlaysEnabled(false);
	//pViewportSmall->setSkiesEnabled(false);
	//pViewportSmall->setShadowsEnabled(false);
}

void createFrameListener()
{
    g_frameListnr = new HRFrameListener(g_pWindow, g_pCamera);
    Root::getSingleton().addFrameListener(g_frameListnr);
}

void createCameraLight()
{
	// Create a camera light (just in case you run into a dark corner :)
	Light* pLight = g_pSceneMgr->createLight("CamLight");
	pLight->setPosition(0.0f, 0.0f, 0.0f); 
	pLight->setAttenuation(500.0f, 0.5f, 1.0f, 0.0f);
	// Accept default settings: point light, white diffuse, just set position
	// attach light to a scene node so the PCZSM can handle it properly (zone-wise)
	// IMPORTANT: Lights (just like cameras) MUST be connected to a scene node!
	SceneNode* pLightNode = g_pCamNode->createChildSceneNode("CamLightNode");
	pLightNode->attachObject(pLight); 
}

void loadDotScene()
{
    g_OgreMaxScene.Load("../Media/Scenes/scene3.scene", 
						g_pWindow,
						OgreMax::OgreMaxScene::NO_OPTIONS,
						(PCZSceneManager*)g_pSceneMgr,
						NULL);
	// resolve portal zone pointers
	((PCZSceneManager*)g_pSceneMgr)->connectPortalsToTargetZonesByLocation();
}

void printNodeChildren(FILE* f, Ogre::SceneNode* scnNode, int iIdentLvl)
{
	Ogre::String str;
	for(int i = 0; i < iIdentLvl; i++)
		str += "    ";

	fprintf(f, "%s%s\n", str.c_str(), scnNode->getName().c_str());

	Ogre::SceneNode::ObjectIterator it = scnNode->getAttachedObjectIterator();
	while(it.hasMoreElements())
	{
	   Entity* entity = static_cast<Entity*>(it.getNext());
	   // do whatever there is to do with the entity...
	   fprintf(f, "%s-  \"%s\"\n", str.c_str(), entity->getName().c_str());
	}

	Ogre::SceneNode::ChildNodeIterator childIter = scnNode->getChildIterator();
	while(childIter.hasMoreElements())
	{
		SceneNode* node = (SceneNode*)(childIter.getNext());
		printNodeChildren(f, node, iIdentLvl + 1);
	}
}

void addResourceLocations()
{
	// Load resource paths from config file
	ConfigFile cf;
	cf.load("../resources.cfg");
	// Go through all sections & settings in the file
	ConfigFile::SectionIterator seci = cf.getSectionIterator();
	String secName, typeName, archName;
	while(seci.hasMoreElements())
	{
		secName = seci.peekNextKey();
		ConfigFile::SettingsMultiMap* settings = seci.getNext();
		ConfigFile::SettingsMultiMap::iterator i;
		for(i = settings->begin(); i != settings->end(); ++i)
		{
			typeName = i->first;
			archName = i->second;
			ResourceGroupManager::getSingleton().addResourceLocation(archName, typeName, secName);
		}
	}
}
